#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

import itertools
import logging
import os
import sqlite3

from .common import *
from .unihan import *
from .phonetics import PhoneticUnit


PHONETICS_TABLE = 'phonetics'



def get_db_path(writable=False):
  if writable:
    return writable_data_path(HAPIDB_FILE)
  else:
    db_path = existing_data_path(HAPIDB_FILE)
    # If it doesn't exist then a writable path is required.
    if db_path is None:
      db_path = writable_data_path(HAPIDB_FILE)
    return db_path



def phonetic_system_to_field(name):
  if isinstance(name, PhoneticUnit):
    name = name.__class__.__name__
  name = name.lower()
  for field in ('pinyin', 'wadeGiles', 'yale', 'zhuyin'):
    if name == field.lower():
      return field
  else:
    raise TypeError('unrecognized field for phonetic system: {}'.format(name))


def table_info_to_fields(field_info):
  '''
  Parse table_info PRAGMA results.
  '''
  # field index, name, type, non_null, default, primary key
  for _, name, typ, _, _, is_primary_key in field_info:
    if is_primary_key:
      pkey = ' PRIMARY KEY'
    else:
      pkey = ''

    yield '{} {}{}'.format(name, typ, pkey)




class HaPiDB(object):
  '''
  Database wrapper.
  '''
  def __init__(self, unihan_dir=None, db_path=None, reset=False, check=False):
    if unihan_dir is None:
      unihan_dir = existing_data_path(UNIHAN_DIR)
    self.unihan_dir = unihan_dir

    if db_path is None:
      db_path = get_db_path(writable=reset)
    self.db_path = db_path
    logging.info('database path: {}'.format(db_path))

    if reset:
      try:
        os.remove(db_path)
      except FileNotFoundError:
        pass
      except PermissionError:
        pass
    create = reset or not os.path.exists(db_path)

    conn = sqlite3.connect(
      db_path,
      detect_types=sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES,
      isolation_level=None
    )
    # https://docs.python.org/3/library/sqlite3.html#sqlite3.Row
    conn.row_factory = sqlite3.Row
    self.conn = conn

    unihan_tables = ('Readings', 'Variants')

    if create:
#       with conn:
#         conn.execute('PRAGMA synchronous = OFF')
      self.create_phonetics_table()
      for t in unihan_tables:
        self.create_unihan_table(t)
      with conn:
        conn.execute('VACUUM')
        conn.execute('PRAGMA INTEGRITY_CHECK')
#         conn.execute('PRAGMA synchronous = NORMAL')

    elif check:
      self.check_table(
        PHONETICS_TABLE,
        list(self.phonetics_column_defs(PHONETIC_EQUVALENTS[0]))
      )
      for t in unihan_tables:
        with UnihanFileReader(self.unihan_filepath(t)) as ufr:
          expected_fields = list(self.unihan_column_defs(sorted(ufr.fields)))
        self.check_table(t, expected_fields)



  def unihan_filepath(self, name):
    '''
    Return the file path for the given Unihan file.
    '''
    if not self.unihan_dir:
      raise RuntimeError('no Unihan data found')
    return os.path.join(self.unihan_dir, UNIHAN_FILE_FORMAT.format(name))



  def unihan_column_defs(self, fields):
    yield 'char TEXT PRIMARY KEY'
    for f in fields:
      if f.endswith('_tone'):
        yield '{} INTEGER'.format(f)
      else:
        yield '{} TEXT'.format(f)



  def phonetics_column_defs(self, fields):
    for f in fields:
      if f == 'Pinyin':
        yield 'Pinyin TEXT PRIMARY KEY'
      elif f == 'Zhuyin':
        yield 'Zhuyin TEXT UNIQUE'
      else:
        yield '{} TEXT'.format(f)



  def check_table(self, name, expected_fields):
    with self.conn:
      c = self.conn.cursor()
      c.execute(r'PRAGMA table_info({})'.format(name))
      current_fields = list(table_info_to_fields(c))
      error = len(current_fields) != len(expected_fields) \
      or not all(x == y.replace(' UNIQUE', '', 1) for (x,y) in zip(current_fields, expected_fields))
    if error:
      logging.error('mismatch in table "{}":\n  expected fields:\n    {}\n  detected fields:\n    {}\nYou may need to reset the database (--reset).'.format(name, '\n    '.join(expected_fields), '\n    '.join(current_fields)))
    return error



  def create_phonetics_table(self):
    fields = PHONETIC_EQUVALENTS[0]
    with self.conn:
      c = self.conn.cursor()
      c.execute(r'DROP TABLE IF EXISTS {}'.format(PHONETICS_TABLE))
      c.execute(r'CREATE TABLE {} ({})'.format(
        PHONETICS_TABLE,
        ','.join(self.phonetics_column_defs(fields))
      ))
      insert_statement = 'INSERT INTO {} VALUES ({})'.format(
        PHONETICS_TABLE,
        ','.join('?' for _ in fields)
      )
      c.execute('BEGIN TRANSACTION')
      c.executemany(insert_statement, PHONETIC_EQUVALENTS[1:])
      c.execute('COMMIT')
      row_count = len(PHONETIC_EQUVALENTS[1:])
      values_count = row_count * len(fields)
      logging.info('inserted {:d} rows ({:d} values) into {}'.format(
        row_count, values_count, PHONETICS_TABLE
      ))



  def create_unihan_table(self, name):
    with UnihanFileReader(self.unihan_filepath(name)) as ufr:
      fields = sorted(ufr.fields)
      with self.conn:
        c = self.conn.cursor()
        c.execute(r'DROP TABLE IF EXISTS {}'.format(name))
        c.execute(r'CREATE TABLE {} ({})'.format(
          name,
          ','.join(self.unihan_column_defs(fields))
        ))
        insert_statement = 'INSERT INTO {} VALUES (?,{})'.format(
          name,
          ','.join('?' for _ in fields)
        )
        args = (
          (char, *(values.get(f, None) for f in fields))
          for (char, values) in ufr
        )
        c.execute('BEGIN TRANSACTION')
        c.executemany(insert_statement, args)
        c.execute('COMMIT')
        logging.info('inserted {:d} rows ({:d} values) into {}'.format(
          ufr.row_count, ufr.values_count, name
        ))



  # "where" is a tuple of a clause and its arguments, e.g. ('foo=?', ('bar',))
  def lookup(self, select, where=None, order_by=None, asc=True):
    with self.conn:
      c = self.conn.cursor()
      if where:
        select = '{} WHERE ({})'.format(select, where[0])

      if order_by:
        select = '{} ORDER BY {} ASC'.format(select, ', '.join(order_by))

      logging.debug(select)
      if where:
        c.execute(select, where[1])
      else:
        c.execute(select)

      yield from c





class MandarinDB(HaPiDB):
  def __init__(self, *args, taiwanese=False, **kwargs):
    super(MandarinDB, self).__init__(*args, **kwargs)
    # Select preferred Mandarin pronunciation.
    # See http://unicode.org/reports/tr38/#kMandarin
    # The TW variant is stored in a separate field.
    if taiwanese:
      self.kMandarin = 'kMandarinTW'
    else:
      self.kMandarin = 'kMandarin'



  def get_hanzi_select_query(self, with_phonetics=True, with_variants=False):
    fields = [
      'Readings.char as hanzi',
      'Readings.{pinyin} as pinyin',
      'Readings.{pinyin}_tone as tone',
      'Readings.kDefinition as definition'
    ]
    joins = list()
    if with_phonetics:
      fields.extend((
        '{phontab}.WadeGiles as wadeGiles',
        '{phontab}.Yale as yale',
        '{phontab}.Zhuyin as zhuyin'
      ))
      joins.append('LEFT JOIN {phontab} ON {phontab}.Pinyin = Readings.{pinyin}')
    if with_variants:
      fields.extend((
        'Variants.kTraditionalVariant as traditionalVariant',
        'Variants.kSimplifiedVariant as simplifiedVariant',
        'Variants.kSemanticVariant as semanticVariant',
        'Variants.kSpecializedSemanticVariant as specializedSemanticVariant'
      ))
      joins.append('LEFT JOIN Variants ON Variants.char = Readings.char')
    select = 'SELECT {} FROM Readings'.format(', '.join(fields))
    if joins:
      joins.insert(0, select)
      select = ' '.join(joins)
    return select.format(pinyin=self.kMandarin, phontab=PHONETICS_TABLE)



  def lookup_hanzi_wrapper(self, where=None, order_by=None, **kwargs):
    select = self.get_hanzi_select_query(**kwargs)
    yield from self.lookup(select, where=where, order_by=order_by)



  def lookup_hanzi(self, hzs=None, with_variants=False):
    if hzs:
      if not isinstance(hzs, list):
        hzs = list(hzs)
      where = (
        ' OR '.join('hanzi=?' for _ in hzs),
        hzs
      )
    else:
      where = None
    order_by = ('hanzi',)
    yield from self.lookup_hanzi_wrapper(
      where=where,
      order_by=order_by,
      with_variants=with_variants
    )



  def lookup_phonetics(self, phonetics=None, with_variants=False):
    if phonetics:
      if not isinstance(phonetics, list):
        phonetics = list(phonetics)
      where = (
        ' OR '.join(
          '({field}=? AND tone=?)'.format(
            field=phonetic_system_to_field(phon),
            pinyin=self.kMandarin
          )
          if phon.tone is not None else
          '({field}=?)'.format(
            field=phonetic_system_to_field(phon)
          )
          for phon in phonetics
        ),
        list(itertools.chain.from_iterable(
          (phon.unit, phon.tone) if phon.tone is not None else (phon.unit,)
          for phon in phonetics
        ))
      )
    else:
      where = None
    order_by = ('pinyin', 'tone', 'hanzi')
    yield from self.lookup_hanzi_wrapper(
      where=where,
      order_by=order_by,
      with_variants=with_variants
    )


  def records_to_dict(self, records, key_field):
    return dict((record[key_field], record) for record in records)
