#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

import glob
import logging
import mimetypes
import os
import urllib.parse

from lxml import etree
import colorsysplus

from .common import existing_data_path, update_symlink, get_tone_color_function, TONE_COLOR_SCHEME, DEFAULT_COLOR_SCHEME
from .phonetics import PhoneticUnit, Pinyin, Zhuyin, WadeGiles, Yale, phonetic_system_to_class, SYSTEMS as PHONETIC_SYSTEMS, DEFAULT_SYSTEM as DEFAULT_PHONETIC_SYSTEM
from .db import phonetic_system_to_field



YELLOWBRIDGE_ZI_SEARCH_URL = 'https://www.yellowbridge.com/chinese/charsearch.php?searchChinese=1&zi={}'
AUDIO_SUBPATH = 'dat/audio'



HTML_TEMPLATE = '''<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1 //EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
    <title>HaPi</title>
    <style type="text/css" />
    <script type="text/javascript" />
  </head>
  <body />
</html>
'''

CSS_STYLE = '''
a
{
  text-decoration: none;
}
body
{
  background-color: white;
}
label
{
  font-weight: bold;
}
rt
{
  font-size: inherit;
  text-align: center;
  margin: auto;
  line-height: 1;
}
rt.phonetic
{
  font-size: 0.5em;
}
ruby
{
  margin: 0;
  line-height: 1;
}
textarea
{
  width: 100%;
}
th, td
{
  padding: 0 0.5em;
}
tr:nth-child(even)
{
  background: #f8f8f8;
}
.hanphon
{
  font-size: 2.5em;
}
.hanphon > a
{
  margin: 0 0.2em;
}
.hanzi
{
  font-size: 2.5em;
}
.phonetic
{
  font-size: 1.5em;
  white-space: nowrap;
}
#buttons
{
  position: fixed;
  bottom: 0;
  right: 0;
}
'''


JAVASCRIPT = '''
function toggle_visibility(elems)
{
  for (var i = 0; i < elems.length; i++)
  {
    if (elems[i].style.visibility === 'hidden')
    {
      elems[i].style.visibility = 'visible';
    }
    else
    {
      elems[i].style.visibility = 'hidden';
    }
  }
}

function toggle_ruby_hanzi(cls, parent_id=null)
{
  toggle_visibility(document.getElementsByClassName('ruby_hanzi'));
}

function toggle_rt()
{
  toggle_visibility(document.getElementsByTagName('rt'));
}
'''



# HTML_BUTTONS = '''    <div id="buttons">
#       <button onclick="toggle_ruby_hanzi();">Toggle hanzi</button>
#       <button onclick="toggle_rt();">Toggle pinyin</button>
#     </div>
# '''



def add_buttons(elem):
  '''
  Add buttons for the defined Javascript functions.
  '''
  button = etree.SubElement(elem, 'button', onclick='toggle_ruby_hanzi();')
  button.text = 'toggle hanzi'
  button = etree.SubElement(elem, 'button', onclick='toggle_rt();')
  button.text = 'toggle phonetics'



def phonetic_system_to_column(name):
  if isinstance(name, PhoneticUnit):
    name = name.__name__
  name = name.lower()
  for column in ('Pinyin', 'Wade-Giles', 'Yale', 'Zhuyin'):
    if name == column.replace('-', '').lower():
      return column
  else:
    raise TypeError('unrecognized column for phonetic system: {}'.format(name))



def find(root, path):
  '''
  Look up elements by xpath. This decomposes the path and prefixes a default
  namespace to enable paths such as 'head/script'.
  '''
  ns_map = dict(root.nsmap)
  ns_fmt = 'ns{:d}'
  i = 1
  ns = ns_fmt.format(i)
  while ns in ns_map:
    i += 1
    ns = ns_fmt.format(i)

  path = '/'.join('{}:{}'.format(ns, x) for x in path.split('/'))
  ns_map[ns] = ns_map[None]
  del(ns_map[None])
  return root.xpath(path, namespaces=ns_map)




def append_text(elem, text):
  try:
    try:
      elem[-1].tail += text
    except TypeError:
      elem[-1].tail = text
  except IndexError:
    try:
      elem.text += text
    except TypeError:
      elem.text = text



def strip_whitespace(root):
  # This is equivalent to stringifying the text and then parsing it with
  # parser = etree.XMLParser(remove_blank_text=True)
  # tree = etree.fromstring(string, parser)
  for elem in root.iter():
    if elem.text is not None and not elem.text.strip():
      elem.text = None
    if elem.tail is not None and not elem.tail.strip():
      elem.tail = None



# TODO
# Figure out how to deal with local paths..
def create_audio(local_uris):
  audio = etree.Element('audio', controls="")
  for uri in local_uris:
    server_uri = '/'.join((AUDIO_SUBPATH, os.path.basename(uri)))
    mimetype, _ = mimetypes.guess_type(uri)
    if mimetype:
      source = etree.SubElement(audio, 'source', src=server_uri, type=mimetype)
    else:
      source = etree.SubElement(audio, 'source', src=server_uri)
  append_text(audio, 'This browser does not support audio elements.')
  return audio



def add_style_color(elem, color):
  if color:
    css = 'color:{};'.format(colorsysplus.rgb_to_hex(*color))
    try:
      elem.attrib['style'] += css
    except KeyError:
      elem.attrib['style'] = css



def add_class(elem, cls):
  try:
    elem.attrib['class'] += ' {}'.format(cls)
  except KeyError:
    elem.attrib['class'] = cls



def link(href, content=None, target='_blank', parent=None):
  if parent is None:
    link = etree.Element('a', href=href, target=target)
  else:
    link = etree.SubElement(parent, 'a', href=href, target=target)
  if isinstance(content, str):
    link.text = content
  elif isinstance(content, type(link)):
    link.append(content)
  return link



# def create_html_table(itr, create_color, full_page=True):
def create_table(columns, alignments, rows):
  '''
  columns:
    An iterable of strings for the table headers.

  alignments:
    An iterable of stings for the "align" attributes.

  rows:
    An iterable of records. Each record should be dictionary that maps columns
    to items. Each item should be an etree element or a string.

  headers, alignments and each row much contain the same number of items.
  '''
  table = etree.Element('table')
  header_tr = etree.SubElement(table, 'tr')
  for header, alignment in zip(columns, alignments):
    th = etree.SubElement(header_tr, 'th', align=alignment)
    th.text = header

  for record in rows:
    tr = etree.SubElement(table, 'tr')
    for column, alignment in zip(columns, alignments):
      td = etree.SubElement(tr, 'td', align=alignment)
      item = record[column]
      if isinstance(item, type(tr)):
        td.append(item)
      elif isinstance(item, str):
        td.text = item
      elif not item:
        td.text = None
      else:
        td.text = str(item)
  return table



def add_head_text(root, tag, text, indent=False):
  try:
    elem = find(root, 'head/{}'.format(tag))[0]
  except IndexError:
    raise ValueError('no such element in head: {}'.format(tag))
  else:
    if indent:
      text = '\n'.join('      {}'.format(l) for l in text.strip().split('\n'))
      elem.text = '\n{}\n    '.format(text)
    else:
      elem.text = text



class XHTML(object):
  PHONETIC_COLUMNS = (
    'Pinyin',
    'Zhuyin',
    'Wade-Giles',
    'Yale',
  )
  HEADER_COLUMN = 'Hanzi'
  OTHER_COLUMNS = (
    'Tone',
    'Unicode',
    'Definitions',
    'Variants',
    'Audio'
  )
  def __init__(
    self,
    color_scheme=DEFAULT_COLOR_SCHEME,
    columns=None,
    numeric_tones=False,
    with_audio=True
  ):
    self.get_tone_color = get_tone_color_function(color_scheme)

    self.with_audio = with_audio
    if not columns:
      columns = list(self.get_columns((
        phonetic_system_to_column(DEFAULT_PHONETIC_SYSTEM),
      )))

    self.columns = columns
    self.numeric_tones = numeric_tones

    parser = etree.XMLParser(remove_blank_text=True)
    root = etree.fromstring(HTML_TEMPLATE, parser)
    self.root = root
    add_head_text(root, 'style', CSS_STYLE, indent=True)
    add_head_text(root, 'script', JAVASCRIPT, indent=True)



  def __str__(self):
    tree = etree.ElementTree(self.root)
    return etree.tostring(tree, pretty_print=True).decode()



  def get_audio_uris(self, phon):
    audio_dir = existing_data_path(AUDIO_SUBPATH)
    if not audio_dir:
      return None
    pattern = '{}/{}{:d}.*'.format(audio_dir, phon.unit, phon.tone)
    return glob.glob(pattern)


  def save(self, path):
    dpath = os.path.dirname(path)
    os.makedirs(dpath, exist_ok=True)
    tree = etree.ElementTree(self.root)
    with open(path, 'wb') as f:
      tree.write(f, pretty_print=True)
    logging.info('saved HTML to {}'.format(path))
    audio_dir = existing_data_path(AUDIO_SUBPATH)
    if audio_dir:
      symlink_path = os.path.join(dpath, AUDIO_SUBPATH)
      update_symlink(audio_dir, symlink_path)


  def output(self, target):
    if isinstance(target, str):
       if target == '-':
        print(self)
       else:
        self.save(target)
    else:
      tree = etree.ElementTree(self.root)
      tree.write(target, pretty_print=True)




  def get_columns(self, phonetics):
    yield self.HEADER_COLUMN
    for p in phonetics:
      yield phonetic_system_to_column(p)
    yield from (c for c in self.OTHER_COLUMNS if (self.with_audio or c != 'Audio'))



  def get_hanzi_link(self, hanzi):
    return YELLOWBRIDGE_ZI_SEARCH_URL.format(urllib.parse.quote(hanzi))



  def set_head_text(self, elem, text, linebreaks=True):
    if linebreaks:
      text = '\n      {}\n    '.format(
        '\n      '.join(text.strip().split('\n'))
      )
    find(self.root, 'head/{}'.format(elem))[0].text = text



  def set_title(self, title):
    self.set_head_text('title', title, linebreaks=False)



  def get_column_alignment(self, name):
    if name.title() in ('Definitions', 'Variants', 'Audio'):
      return 'left'
    else:
      return 'center'



  def get_rows(self, records):
    get_tone_color = self.get_tone_color
    for record in records:
      hanzi = record['hanzi']
      tone = record['tone']
      color = get_tone_color(tone)

      row = dict(
        Definitions=record['definition'],
        Tone='{:d}'.format(tone),
        Unicode='U+{:X}'.format(ord(hanzi))
      )

      hanzi_link = link(self.get_hanzi_link(hanzi), content=hanzi)
      add_class(hanzi_link, 'hanzi')
      if color:
        add_style_color(hanzi_link, color)
      row['Hanzi'] = hanzi_link


      phonetic_units = dict()
      for system in ('pinyin', 'zhuyin', 'wadegiles', 'yale'):
        field = phonetic_system_to_field(system)
        pu_cls = phonetic_system_to_class(system)
        column = phonetic_system_to_column(system)

        unit = record[field]
        if unit is None:
          text = '?'
        else:
          pu = pu_cls((unit, tone), numeric=self.numeric_tones)
          phonetic_units[column] = pu
          text = pu.with_tone_mark()

        span = etree.Element('span')
        span.text = text
        add_class(span, 'phonetic')
        if color:
          add_style_color(span, color)
        row[column] = span

      variants_table = etree.Element('table')
      for header, key in (
        ('traditional', 'traditionalVariant'),
        ('simplified', 'simplifiedVariant'),
        ('semantic', 'semanticVariant'),
        # There is a non-breaking space here (U+00A0).
        ('specialized semantic', 'specializedSemanticVariant')
      ):
        value = record[key]
        if not value:
          continue
        tr = etree.SubElement(variants_table, 'tr')
        th = etree.SubElement(tr, 'th', align='left')
        th.text = header
        td = etree.SubElement(tr, 'td')
        for variant in value.split():
          varchar = variant[0]
          varqual = variant[1:]
          var_link = link(self.get_hanzi_link(varchar), content=varchar, parent=td)
          if varqual:
            var_link.tail = varqual
      if len(variants_table):
        row['Variants'] = variants_table
      else:
        row['Variants'] = ''

      row['Audio'] = ''
      if self.with_audio:
        audio_uris = self.get_audio_uris(phonetic_units['Pinyin'])
        if audio_uris:
          row['Audio'] = create_audio(audio_uris)

      yield row



  def create_table(self, rows, columns=None):
    if not columns:
      columns = self.columns
    alignments = tuple(self.get_column_alignment(c) for c in columns)
    return create_table(columns, alignments, rows)



  def create_button_panel(self):
    div = etree.Element('div')
    add_buttons(div)
    return div



  def create_paragraphs(self, text, hanzi_map, phonetics=None):
    get_tone_color = self.get_tone_color
    line_height = 2

    if phonetics:
      if isinstance(phonetics, str):
        phonetics = (phonetics,)
      line_height += 0.6 * len(phonetics)
      phonetics = tuple(p.lower() for p in phonetics)
      p_cls = 'hanphon'
      systems = dict(
        (
          p,
          (
            phonetic_system_to_class(p),
            phonetic_system_to_field(p)
          )
        ) for p in phonetics
      )
    else:
      p_cls = 'hanzi'

    p = etree.Element('p')
    add_class(p, p_cls)
    line_height_style = 'line-height: {:f};'.format(line_height)
    p.attrib['style'] = line_height_style

    for hz in text:
      try:
        record = hanzi_map[hz]
      except KeyError:
        if hz == '\n':
          # Don't yield empty elements.
          if len(p) > 0:
            yield p
            p = etree.Element('p')
            add_class(p, p_cls)
            p.attrib['style'] = line_height_style
        else:
          append_text(p, hz)
        continue
      else:
        tone = record['tone']
        color = get_tone_color(tone)

#         details = etree.SubElement(p, 'details')
#         summary = etree.SubElement(details, 'summary')
#         summary.text = hz
#         if phonetics:
#           for i, system in enumerate(phonetics):
#             pu_cls, field = systems[system]
#             dp = etree.SubElement(details, 'p')
#             dp.text = pu_cls((record[field], tone)).with_tone_mark()

#         block = etree.SubElement(p, 'span')
#         add_style_color(block, color)
#         add_class(block, 'hanziblock')
#         elem = etree.SubElement(block, 'span')
#         elem.text = hz
#         if phonetics:
#           for system in phonetics:
#             pu_cls, field = systems[system]
#             elem = etree.SubElement(block, 'span')
#             elem.text = pu_cls((record[field], tone)).with_tone_mark()
#             add_class(elem, 'rt_{}'.format(system))
#             add_class(elem, 'phonetics')



        hanzi_link = link(self.get_hanzi_link(hz), parent=p)
        ruby = etree.SubElement(hanzi_link, 'ruby')
        span = etree.SubElement(ruby, 'span')
        add_class(span, 'ruby_hanzi')
        span.text = hz
        add_style_color(ruby, color)

        # Abuse nested ruby elements to align multiple phonetic annotations.
        if phonetics:

          def ruby_miner(elem, phonetics):
            rt = etree.SubElement(elem, 'rt')
            system = phonetics[0]
            pu_cls, field = systems[system]
            txt = pu_cls((record[field], tone), numeric=self.numeric_tones).with_tone_mark()

            phonetics = phonetics[1:]

            if phonetics:
              rb = etree.SubElement(rt, 'ruby')
              rb.text = txt
              ruby_miner(rb, phonetics)
            else:
              rt.text = txt
            return rt

          rt = ruby_miner(ruby, phonetics)
          add_class(rt, 'phonetic')
          # This ensures that an invisible space is present between the phonetic
          # units when copying them from the page.
#           span = etree.SubElement(rt, 'span', style='font-size:0;')
#           span.text = ' '

    if len(p) > 0:
      yield p


  def create_text_table(self, hanzi_map, sort_by=None, columns=None):
    if sort_by:
      records = (record for (_, record) in sorted(
        hanzi_map.items(),
        key=lambda x: x[1][sort_by]
      ))
    else:
      records = (record for (_, record) in sorted(hanzi_map.items()))
    return self.create_table(self.get_rows(records), columns=columns)


  def add_inputs(self, elem, label, typ, name, values, check=None):
    if isinstance(check, str):
      check = (check,)
    n = len(values)
    w = len(str(n))
    name_fmt = 'list-{{}}-{{:0{:d}d}}'.format(w)
    etree.SubElement(elem, 'label').text = label
    for i,v in enumerate(values):
      if typ == 'checkbox' and n > 1:
        this_name = name_fmt.format(name, i)
      else:
        this_name = name
      checked = (check and v in check)
      if v is None:
        val = ''
        v = 'none'
      else:
        val = v
      ipt = etree.SubElement(elem, 'input', type=typ, name=this_name, value=val)
      if n > 1:
        ipt.tail = v
      if checked:
        ipt.attrib['checked'] = '1'
    etree.SubElement(elem, 'br')


  def create_form(self, action='/output.htm'):
    form = etree.Element('form', action=action, method='post', target='_blank')

    self.add_inputs(
      form, 'Color Scheme', 'radio', 'color',
      (None, *TONE_COLOR_SCHEME),
      check=DEFAULT_COLOR_SCHEME
    )
    self.add_inputs(
      form, 'Phonetic System', 'checkbox', 'phonetics',
      PHONETIC_SYSTEMS,
      check=DEFAULT_PHONETIC_SYSTEM
    )
    self.add_inputs(
      form, 'Numeric Tones', 'checkbox', 'numeric_tones',
      ('numeric_tones',)
    )
    self.add_inputs(
      form, 'Taiwanese', 'checkbox', 'taiwanese',
      ('taiwanese',)
    )
    self.add_inputs(
      form, 'Include Table', 'checkbox', 'include_table',
      ('include_table',),
      check='include_table'
    )
    self.add_inputs(
      form, 'Include Audio', 'checkbox', 'include_audio',
      ('include_audio',),
      check='include_audio'
    )

    input_label = etree.SubElement(form, 'label')
    input_label.attrib['for'] = 'input'
    input_label.text = 'Input'
    etree.SubElement(form, 'br')
    textarea = etree.SubElement(form, 'textarea', name='input', id='input', rows='15', cols='120')
    textarea.text = ''
    etree.SubElement(form, 'br')

    submit = etree.SubElement(form, 'input', type='submit', value='Submit')

    return form


  def append_section(self, elems, title=None, level=1):
    body = find(self.root, 'body')[0]
    if title:
      h = etree.SubElement(body, 'h{:d}'.format(level))
      h.text = title
    if isinstance(elems, type(body)):
      body.append(elems)
    else:
      for elem in elems:
        if isinstance(elem, str):
          append_text(body, elem)
        else:
          body.append(elem)


  def set_title(self, title):
    add_head_text(self.root, 'title', title)



  def clear_body(self):
    find(self.root, 'body')[0].clear()
