#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

import logging

from .common import PHONETIC_EQUVALENTS

SYSTEMS = ('pinyin', 'zhuyin', 'wadeGiles', 'yale')
DEFAULT_SYSTEM = 'pinyin'
PLACEHOLDER = '<N/A>'

def phonetic_system_to_class(system):
  system = system.lower()
  if system == 'pinyin':
    return Pinyin
  elif system == 'zhuyin':
    return Zhuyin
  elif system == 'wadegiles':
    return WadeGiles
  elif system == 'yale':
    return Yale
  else:
    raise ValueError('unrecognized phonetic system: {}'.format(system))



def deduce_unit(word):
  units = tuple(phonetic_system_to_class(s)(word) for s in PHONETIC_EQUVALENTS[0])
  for alternatives in PHONETIC_EQUVALENTS[1:]:
    for unit, alt in zip(units, alternatives):
      if unit.unit == alt:
        return unit
  return units[0]



class PhoneticUnit(object):
  '''
  Class to represent different phonetic units (e.g. Pinyin, Zhuyin).
  '''
  TONE_RANGE = tuple()

  def __init__(self, unit, numeric=False, default_tone=None):
    '''
    unit:
      The word or symbol that represents the sound. If it is a string then it
      will be parsed to determine the tone. If it is a 2-tuple, the first item
      will be parsed as the unit and the second as the tone.

    numeric:
      If True, prefer numeric tone marks when the phonetic system supports them.
    '''
    if isinstance(unit, str):
      self.unit, self.tone = self.parse(unit)
      if self.tone is None and default_tone:
        if default_tone is True:
          self.tone = self.NEUTRAL_TONE
        else:
          self.tone = default_tone
    else:
      unit, self.tone = unit
      # Parse here to ensure that the unit has been stripped of tone marks.
      self.unit, _ = self.parse(unit)
    self.numeric = numeric

  def parse(self, unit):
    '''
    Parse tone marks and return the unmarked unit along with the parsed numeric
    tone.
    '''
    return NotImplemented, NotImplemented

  def with_tone_mark(self, tone=None):
    '''
    Add the tone mark to the unit.

    For systems such as Pinyin that support both diacritics and numbers, this
    should check self.numeric and return the appropriate value.
    '''
    return PLACEHOLDER

  def __iter__(self):
    '''
    Return an iterable over tone-marked units. If the tone is None, return an
    iterator over all possible toned units, otherwise return a 1-tuple with the
    tone.
    '''
    if self.tone is not None:
      return (self.with_tone_mark(),)
    else:
      return (self.with_tone_mark(tone=t) for t in self.TONE_RANGE)



class Mandarin(PhoneticUnit):
  TONE_RANGE = (1,2,3,4,5)
  NEUTRAL_TONE = 5



class Pinyin(Mandarin):
  # AltGr combinations for the tone marks:
  # * tone 1: #
  # * tone 2: '
  # * tone 3: >
  # * tone 4: `
  # " for ü, tone marks + v for ǖ, ǘ, ǚ and ǜ
  #
  # Unihan also includes the following: hm, n, ń, ň, ǹ, ḿ
  # For now they are ignored as I do not yet know what to do with them.
  #
  # Fifth tone is at string index 0, the other tones match their index.
  TONE_MARKS = ' ¯´ˇ`'
  MARKED = {
    'a': 'aāáǎà',
    'e': 'eēéěè',
    'i': 'iīíǐì',
    'o': 'oōóǒò',
    'u': 'uūúǔù',
    'ü': 'üǖǘǚǜ'
  }

  UNMARKED = dict(
    (m, (unmarked, tone))
    for unmarked, marked in MARKED.items()
    for tone, m in enumerate(marked)
  )

  def parse(self, unit):
    if not unit:
      return None, None
    tone = None
    unmarked_map = self.UNMARKED

    if unit[-1].isdigit():
      unit, tone = unit[:-1], int(unit[-1])

    else:
      for c in unit:
        try:
          unmarked, maybe_tone = unmarked_map[c]
          # pinyin such as "nián" will return the neutral tone on "i". Keep going
          # to find a possibly toned character.
          if not maybe_tone:
            raise KeyError
        except KeyError:
          continue
        else:
          unit = unit.replace(c, unmarked, 1)
          tone = maybe_tone
          break

    unit = unit.replace('ü', 'v', 1)
    return unit, tone



  # http://pinyin.info/rules/where.html
  def with_tone_mark(self, tone=None):
    '''
    Add the tone mark to the word.
    '''
    if not self.unit:
      return PLACEHOLDER
    word = self.unit.replace('v', 'ü', 1)
    if tone is None:
      tone = self.tone
    marked = self.MARKED

    if self.numeric:
      return '{}{:d}'.format(word, tone)

    try:
      tone %= 5
    except TypeError:
      tone = 0
    if not tone:
      return word
    if 'a' in word:
      word = word.replace('a', marked['a'][tone], 1)
    elif 'e' in word:
      word = word.replace('e', marked['e'][tone], 1)
    elif 'ou' in word:
      word = word.replace('o', marked['o'][tone], 1)
    # First vowel gets the tone mark
    else:
      for c in word:
        try:
          word = word.replace(c, marked[c][tone], 1)
          break
        except KeyError:
          continue
      # This is an error for pinyin but the Yale romanization includes units
      # without vowels such as "shr" and "dz".
      # TODO
      # Determine how to correctly handle such units.
      else:
        word += self.TONE_MARKS[tone]
#         logging.error('failed to add tone mark to {} [tone {:d}]'.format(word, tone))
    return word



class Zhuyin(Mandarin):
  # Neutral fifth tone is at index 0. Other tones match their index.
  MARKS = ['˙', '', 'ˊ', 'ˇ', 'ˋ']

  def parse(self, unit):
    if not unit:
      return None, None
    marks = self.MARKS
    if unit[-1] in marks:
      return unit[:-1], marks.index(unit[-1])
    else:
      return unit, None


  def with_tone_mark(self, tone=None):
    if not self.unit:
      return PLACEHOLDER
    if tone is None:
      tone = self.tone
    if tone is None:
      return self.unit
    elif self.numeric:
      return '{}{:d}'.format(self.unit, tone)
    else:
      return '{}{}'.format(self.unit, self.MARKS[tone % 5])



class WadeGiles(Zhuyin):
  # Neutral fifth tone is at index 0. Other tones match their index.
  MARKS = ['', '¹', '²', '³', '⁴']


# This needs to be checked for tone parsing.
class Yale(Pinyin):
  pass


