#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

from http.server import HTTPServer, BaseHTTPRequestHandler
import argparse
import html
import logging
import mimetypes
import os
import time
import urllib.parse

from .db import MandarinDB
from .common import configure_logging, DEFAULT_COLOR_SCHEME, existing_data_path
from .html import XHTML, AUDIO_SUBPATH



############################# From ThreadedServers #############################

DEFAULT_CHUNK_SIZE = 0x1000
DEFAULT_CHUNK_FACTOR = 0x10 #0x80

def format_size(size):
  """Format bytes for inferior humans."""
  if size < 0x400:
    return '{:d} B'.format(size)
  else:
    size = float(size) / 0x400
  for prefix in ('KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB'):
    if size < 0x400:
      return '{:0.02f} {}'.format(size, prefix)
    else:
      size /= 0x400
  return '{:0.02f} YiB'.format(size)



def get_valid_ranges(ranges, size):
  for s, e in ranges:
    s = max(0, s)
    e = min(size, e)
    if s >= e \
    or e < 0 \
    or s >= size:
      continue
    yield s, e



def parse_range_header(header, size=None):
  '''
  Parse HTTP "Range" headers. Range headers may take the form "a-b,c-d,e-f".

  This function assumed that the range headers are well-formed, i.e. that there
  are no extra spaces, invalid offsets, or extra characters.
  '''
  try:
    unit, ranges = header.split('=', 1)
  except ValueError:
    raise ValueError('Range header does not contain "="')

  if unit != 'bytes':
    raise ValueError('unrecognized Range unit: {}'.format(unit))

  for r in ranges.split(','):
    start, end = r.split('-')
    if start:
      try:
        start = int(start)
      except ValueError:
        raise ValueError('invalid start in Range header: {}'.format(start))
    else:
      start = 0
    if end:
      try:
        end = int(end) + 1
      except ValueError:
        raise ValueError('invalid end in Range header: {}'.format(end))
    else:
      end = size
    yield start, end



def format_content_range(ranges, size):
  '''
  Format "Content-Range" headers.
  '''
  return 'bytes {}/{:d}'.format(
    ','.join('{:d}-{:d}'.format(s,e-1) for s,e in ranges),
    size,
  )



def get_length_from_ranges(ranges):
  '''
  Calculate the total content length from a list of ranges.
  '''
  n = 0
  for s, e in ranges:
    n += e - s
  return n

################################################################################



class HapiRequestHandler(BaseHTTPRequestHandler):

  # Overrite the log timestamp format.
  def log_date_time_string(self):
    return time.strftime('%Y-%m-%d %H:%M:%S %Z')



  # Override to use the logging module.
  def log_message(self, fmt, *args):
    if args:
      string = fmt % args
    else:
      string = fmt
    msg = '{} {}'.format(self.client_address[0], string)
    logging.info(msg)



  # Override to use the logging module.
  def log_error(self, fmt, *args):
    if args:
      string = fmt % args
    else:
      string = fmt
    msg = '{} {}'.format(self.client_address[0], string)
    logging.error(msg)



  # From ThreadedServers.HTTPS
  def hapi_transfer_file(
    self,
    fpath,
    include_body=True,
    chunk_size=None,
  ):
    '''
    Transfer the contents of a file, with support for the Range header.
    '''
    size = os.path.getsize(fpath)
    mimetype, encoding = mimetypes.guess_type(fpath)
    mtime = os.path.getmtime(fpath)
    last_modified = self.date_time_string(mtime) # RFC 2822
    if not mimetype:
      mimetype = 'application/octet-stream'

    if 'Range' in self.headers:
      try:
        ranges = list(
          get_valid_ranges(
            parse_range_header(self.headers['Range'], size=size),
            size
          )
        )
        code = 206
      except ValueError:
        self.send_error(400)
        return
      if not ranges:
        self.send_error(416) # Requested Range Not Satisfiable
        return
    else:
      ranges = list(((0, size),))
      code = 200

    if len(ranges) > 1:
      mimetype = 'multipart/byteranges'

    if 'If-Modified-Since' in self.headers:
      if_modifed_since = time.strptime(
        self.headers['If-Modified-Since'], RFC_2822_TIME_FORMAT
      )
      if mtime <= calendar.timegm(if_modifed_since):
        code = 304
        include_body = False

    self.send_response(code)
    self.send_header('Content-Type', mimetype)
    self.send_header('Last-Modified', last_modified)
    self.send_header('Accept-Ranges', 'bytes')

    if include_body:
      self.send_header('Content-Length', get_length_from_ranges(ranges))
      if encoding:
        self.send_header('Content-Encoding', encoding)
      self.send_header('Content-Range', format_content_range(ranges, size))
      self.end_headers()

      # This does not affect "include_body" above because a HEAD request should
      # return the same metadata as a GET request.
      if self.command != 'HEAD':
        try:
          with open(fpath, 'rb') as f:
            if chunk_size is None:
              chunk_size = getattr(f, '_CHUNK_SIZE', DEFAULT_CHUNK_SIZE)
            chunk_size *= DEFAULT_CHUNK_FACTOR

            ranges_sent = list()
            for start, end in ranges:
              f.seek(start, os.SEEK_SET)
              remaining = end - start
              bytes_sent = 0
              while remaining > 0:
                chunk = f.read(min(remaining, chunk_size))
                if chunk:
                  bytes = self.wfile.write(chunk)
                  bytes_sent += bytes
#                   bytes = len(chunk) #Python2
                  remaining -= bytes
                  if bytes != len(chunk): #Python3
                    break #Python3
                else:
                  self.log_error('local EOF error [{}]'.format(fpath))
              ranges_sent.append((start, end, bytes_sent))

          # Report file transfer data.
          total_bytes_sent = 0
          number_of_ranges = len(ranges)
          report_total = number_of_ranges > 1

          for start, end, bytes_sent in ranges_sent:
            range_size = end - start
            total_bytes_sent += bytes_sent

            if range_size == size:
              self.log_message(
                '{} / {} transferred [{}]'.format(
                  format_size(bytes_sent),
                  format_size(size),
                  fpath
                )
              )
            else:
              self.log_message(
                '{} / {} transferred [range {:d}-{:d}/{:d} of {}]'.format(
                  format_size(bytes_sent),
                  format_size(range_size),
                  start,
                  end,
                  size,
                  fpath
                )
              )

          if report_total:
            self.log_message(
              '{} / {} transferred in {:d} ranges [{}]'.format(
                format_size(total_bytes_sent),
                size,
                number_of_ranges,
                fpath
              )
            )


        except (BrokenPipeError, ConnectionResetError) as e:
            user_agent = self.headers['User-Agent']
            if user_agent is None:
              user_agent = ''
            else:
              user_agent = ' [{}]'.format(user_agent)
            self.log_error('connection closed by client{}'.format(user_agent))
    else:
      self.send_header('Content-Length', '0')
      self.end_headers()



  def hapi_serve_xhtml(self, xhtml):
    self.send_response(200, 'OK')
    self.send_header('Content-Type', 'application/xhtml+xml')
    self.send_header('Connection', 'close')
    self.end_headers()
    xhtml.output(self.wfile)



  def hapi_do_UNHANDLED(self):
    self.send_error(405)
    self.send_header('Connection', 'close')
    self.end_headers()



  def do_HEAD(self):
    return self.hapi_do_UNHANDLED()



  def do_PUT(self):
    return self.hapi_do_UNHANDLED()



  def do_GET(self):
    if self.path.startswith('/{}'.format(AUDIO_SUBPATH)):
      local_path = existing_data_path(self.path[1:])
      if not local_path:
        return self.hapi_do_UNHANDLED()
      else:
        return self.hapi_transfer_file(local_path)

    else:
      xhtml = XHTML()
      form = xhtml.create_form()
      xhtml.append_section(form)
      return self.hapi_serve_xhtml(xhtml)



  def do_POST(self):
    content_len = int(self.headers.get('content-length', 0))
    content = self.rfile.read(content_len)
    data = dict(
      (k.decode(), html.unescape(v.decode()))
      for (k,v) in (field.split(b'=', 1) for field in content.split(b'&'))
    )
    data_lists = dict()
    for (k,v) in sorted(data.items()):
      if k.startswith('list'):
        name = k.split('-', 2)[1]
        try:
          data_lists[name].append(v)
        except KeyError:
          data_lists[name] = [v]
        del data[k]
    data.update(data_lists)

    text = urllib.parse.unquote_plus(data.get('input', ''))
    hapidb = MandarinDB(taiwanese=bool(data.get('taiwansese')))
    records = hapidb.lookup_hanzi(text, with_variants=True)
    hanzi_map = hapidb.records_to_dict(records, 'hanzi')

    xhtml = XHTML(
      color_scheme=data.get('color', DEFAULT_COLOR_SCHEME),
      numeric_tones=bool(data.get('numeric_tones')),
      with_audio=bool(data.get('include_audio'))
    )

    phonetics = data.get('phonetics')
    if phonetics:
      xhtml.columns = list(xhtml.get_columns(phonetics))


    buttons = xhtml.create_button_panel()
    xhtml.append_section(buttons)
    paragraphs = xhtml.create_paragraphs(text, hanzi_map, phonetics=phonetics)
    xhtml.append_section(paragraphs)
    if data.get('include_table'):
      table = xhtml.create_text_table(hanzi_map, sort_by='Pinyin')
      xhtml.append_section(table)
    xhtml.set_title('Colorized Hanzi Text')

    return self.hapi_serve_xhtml(xhtml)



def get_parser():
  parser = argparse.ArgumentParser(
    description='Run a local HaPi server to input text via your browser.',
  )
  parser.add_argument(
    '-a', '--address', default='localhost',
    help='The server address. Pass en empty string to listen on all interfaces. Default: %(default)s'
  )
  parser.add_argument(
    '-p', '--port', type=int, default=8000,
    help='The server port. Default: %(default)s'
  )
  return parser



def run(address, port):
  httpd = HTTPServer((address, port), HapiRequestHandler)
  logging.info('starting HTTP server [address: "{}", port: {:d}] (press ctrl+c to exit)'.format(address, port))
  httpd.serve_forever()


if __name__ == '__main__':
  configure_logging()
  parser = get_parser()
  pargs = parser.parse_args()
  try:
    run(pargs.address, pargs.port)
  except KeyboardInterrupt:
    pass
