#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

import os
import re

from .phonetics import Pinyin

def insert_characters(text):
  '''
  Replace Unicode escapes with characters.
  '''
  return re.sub(
    r'[uU]\+([0-9A-Fa-f]+)',
    lambda x: chr(int(x.group(1), 16)),
    text
  )



def parse_unihan_field_names(f):
  '''
  Parse the names of the fields in the header of a Unihan file and return the
  file object positioned at the beginning of the data.
  '''
  fields = set()
  last_line = 0
  # Iterating over f disables tell().
  f.seek(0, os.SEEK_SET)
  while True:
    line = f.readline()
    if not line.startswith('#'):
      break
    elif line.startswith('#\tk'):
      field = line[2:-1]
      if field == 'kMandarin':
        fields.add('kMandarin')
        fields.add('kMandarin_tone')
        fields.add('kMandarinTW')
        fields.add('kMandarinTW_tone')
      else:
        fields.add(field)
    last_line = f.tell()
  f.seek(last_line, os.SEEK_SET)
  return fields



def parse_unihan_data(f):
  '''
  Collect fields by character and yield them together with the character.
  '''
  last_char = None
  fields = dict()
  for line in f:
    line = line.strip()
    if not line or line[0] == '#':
      continue
    codepoint, field, value = line.split(None, 2)
    char = chr(int(codepoint[2:], 16))
    if char != last_char:
      if fields:
        yield last_char, fields
        fields.clear()
    last_char = char

    if field == 'kMandarin':
      # If two values are given, the first is preferred for zh-Hans (CN) and the
      # second is preferred for zh-Hant (TW).
      values = value.split()
      py = Pinyin(values[0], default_tone=True)
      fields['kMandarin'], fields['kMandarin_tone'] = py.unit, py.tone
      try:
        py = Pinyin(values[1], default_tone=True)
      except IndexError:
        pass
      fields['kMandarinTW'], fields['kMandarinTW_tone'] = py.unit, py.tone
      continue
    elif field != 'kDefinition':
      value = re.sub(r'U\+([0-9A-F]+)', lambda x: chr(int(x.group(1), 16)), value)
    fields[field] = value
  if fields:
    yield last_char, fields



class UnihanFileReader(object):
  '''
  Context manager for reading Unihan files.
  '''
  def __init__(self, path):
    self.fileobject = None
    self.path = path
    self.fields = None
    self.rows = None
    self.row_count = 0
    self.values_count = 0

  def __enter__(self):
    self.fileobject = f = open(self.path, 'r')
    self.fields = parse_unihan_field_names(f)
    self.rows = parse_unihan_data(f)
    self.row_count = 0
    self.values_count = 0
    return self

  def __exit__(self, type, value, traceback):
    self.fileobject.close()

  def __iter__(self):
    return self

  def __next__(self):
    '''
    Iterate while counting values for logging.
    '''
    char, values = next(self.rows)
    self.row_count += 1
    self.values_count += len(values)
    return char, values